/*
 */
package de.pfeifer_syscon.corvid;

import COSE.Message;
import COSE.OneKey;
import COSE.Sign1Message;
import com.upokecenter.cbor.CBORObject;
import java.util.ArrayList;
import java.util.List;
import java.util.zip.Inflater;
import nl.minvws.encoding.Base45;

/**
 *
 * @author rpf
 */
public class CertMessage implements CertInterface {
    private static final int BUFFER_SIZE = 256;   // use smaller fragments to extract
    private static final int COUNTRY_TAG = 1;
    private static final int KID_TAG = 4;
    private static final int  HEALTH_CERTIFICATE_CLAIM = -260;
    private static final int  DIGITAL_GREEN_CERTIFICATE = 1;
    private static final String VACCINATION_TAG = "v";
    private static final String NAME_TAG = "nam";
    private static final String DATEOFBIRTH_TAG = "dob";
    private static final String VERSION_TAG = "ver";
    private static final String FAMILYNAME_TAG = "fn";
    private static final String GIVENNAME_TAG = "gn";
    private static final String DOSENUM_TAG = "dn";
    private static final String TOTALSERIALDOSE_TAG = "sd";
    private static final String DATE_TAG = "dt";
    private static final String MANUFACTURER_TAG = "ma";
    private static final String VACCCINECODE_TAG = "vp";
    private static final String VACCINATION_COUNTRY_TAG = "co";
    private static final String CERTINFO_TAG = "ci";
    private static final String ISSUER_TAG = "is";
    private static final String MEDICALPRODUCT_TAG = "mp";
    private static final String TARGETEDDISEASE_TAG = "tg";

    private final String text;
    private final CBORObject cborMap;

    public CertMessage(String text) throws Exception {
        this.text = text;
        if (text == null || text.trim().length() == 0) {
            throw new IllegalArgumentException("The CertMessage text shoud not be empty!");
        }
        if (!text.startsWith("HC1:")) {
            throw new IllegalArgumentException("No HC1: prefix from qr text!!!");
        }
        String withoutPrefix = text.substring(4);
        byte[] bytecompressed = Base45.getDecoder().decode(withoutPrefix);
        byte[] min = decompress(bytecompressed);
        Message a = Sign1Message.DecodeFromBytes(min);
        if (!(a instanceof Sign1Message)) {
            throw new Exception(String.format("Not a sign1 message but %s !!!", a.getClass().getSimpleName()));
        }
        Sign1Message signMessage = (Sign1Message)a;
        CBORObject cborMapUnprot = signMessage.getUnprotectedAttributes();
        CBORObject cborMapProt = signMessage.getProtectedAttributes();

        byte[] content = a.GetContent();
        cborMap = CBORObject.DecodeFromBytes(content);

        // if the following is uncommented validation shoud fail...
        //CBORObject nameTag = getNameTag();
        //nameTag.set(CBORObject.FromObject(FIRSTNAME_TAG), CBORObject.FromObject("Xxxxxx"));
        //signMessage.SetContent(cborMap.EncodeToBytes());

        String country = cborMap.get(CBORObject.FromObject(COUNTRY_TAG)).AsString();
        byte[] kid = cborMapUnprot.get(CBORObject.FromObject(KID_TAG)).GetByteString();
        TrustedKey matchKey = TrustedKeys.find(country, kid);
        OneKey key = matchKey.getKey();
        if (!signMessage.validate(key)) {
            throw new Exception("Message cound not be validated!!!");
        }
    }

    @Override
    public String getRawString() {
        return text;
    }



    /**
     * @return a text repesentation of corvid certificate
     */
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(String.format("Name %s\n", getName()));
        sb.append(String.format("DateOfBirth %s\n", getDateOfBirth()));
        sb.append(String.format("%s\n", getVaccination()));
        sb.append(String.format("Version %s", getVersion()));

        return sb.toString();
    }

    /**
     * @return a json repesentation of corvid certificate
     */
    public String getJsonString() {
        return cborMap.ToJSONString();
    }

    public CBORObject getHealthClaim() {
        return cborMap.get(CBORObject.FromObject(HEALTH_CERTIFICATE_CLAIM));
    }

    public CBORObject getGreenCert() {
        return getHealthClaim().get(CBORObject.FromObject(DIGITAL_GREEN_CERTIFICATE));
    }

    public CBORObject getVaccinationTag() {
        return getGreenCert().get(CBORObject.FromObject(VACCINATION_TAG));
    }

    interface VaccinationExtractor {
        String extract(CBORObject vacTag);
    }

    private String vaccinationExtractor(VaccinationExtractor vaccinationExtractor) {
        CBORObject vacTags = getVaccinationTag();
        StringBuilder sb = new StringBuilder();
        for (CBORObject vacTag : vacTags.getValues()) {
            if (sb.length() > 0) {
                sb.append(", ");
            }
            String ci = vaccinationExtractor.extract(vacTag);
            sb.append(ci);
        }
        return sb.toString();
    }


    @Override
    public String getCertIdentifier() {
        return vaccinationExtractor((CBORObject vacTag) -> {
            return vacTag.get(CBORObject.FromObject(CERTINFO_TAG)).AsString();
        });
    }

    @Override
    public String getVaccination() {
        CBORObject vacTags = getVaccinationTag();
        StringBuilder sb = new StringBuilder();
        for (CBORObject vacTag : vacTags.getValues()) {
            String date = vacTag.get(CBORObject.FromObject(DATE_TAG)).AsString();
            sb.append(String.format("Date %s\n", date));
            int dn = vacTag.get(CBORObject.FromObject(DOSENUM_TAG)).AsInt32();
            int sd = vacTag.get(CBORObject.FromObject(TOTALSERIALDOSE_TAG)).AsInt32();
            sb.append(String.format("Dose %d/%d\n", dn, sd));
            String targeted = vacTag.get(CBORObject.FromObject(TARGETEDDISEASE_TAG)).AsString();
            sb.append(String.format("TargetDisease %s\n", targeted));
            String product = vacTag.get(CBORObject.FromObject(MEDICALPRODUCT_TAG)).AsString();
            sb.append(String.format("Product %s\n", product));
            String manufacturer = vacTag.get(CBORObject.FromObject(MANUFACTURER_TAG)).AsString();
            sb.append(String.format("Manufacturer %s\n", manufacturer));
            String vp = vacTag.get(CBORObject.FromObject(VACCCINECODE_TAG)).AsString();
            sb.append(String.format("VacccineCode %s\n", vp));
            String ci = vacTag.get(CBORObject.FromObject(CERTINFO_TAG)).AsString();
            sb.append(String.format("CertInfo %s\n", ci));
            String issuer = vacTag.get(CBORObject.FromObject(ISSUER_TAG)).AsString();
            sb.append(String.format("Issuer %s\n", issuer));
            String country = vacTag.get(CBORObject.FromObject(VACCINATION_COUNTRY_TAG)).AsString();
            sb.append(String.format("Country %s", country));
        }
        return sb.toString();
    }

    // the following functions just decode some infos, if you need more have
    //   a look at covpass-android:vocpass-sdk EUValueSetRepository
    @Override
    public String getTargeted() {
        return vaccinationExtractor((CBORObject vacTag) -> {
            String targeted = vacTag.get(CBORObject.FromObject(TARGETEDDISEASE_TAG)).AsString();
            switch (targeted)  {
            case "840539006":
                return "COVID-19";
            }
            return targeted;
        });
    }

    @Override
    public String getVaccine() {
        return vaccinationExtractor((CBORObject vacTag) -> {
            String vaccine = vacTag.get(CBORObject.FromObject(VACCCINECODE_TAG)).AsString();
            switch (vaccine)  {
            case "1119349007":
                return "SARS-CoV-2 mRNA vaccine";
            }
            return vaccine;
        });
    }

    @Override
    public String getProduct() {
        return vaccinationExtractor((CBORObject vacTag) -> {
            String product = vacTag.get(CBORObject.FromObject(MEDICALPRODUCT_TAG)).AsString();
            switch (product)  {
            case "EU/1/20/1528":
                return "Comirnaty";
            }
            return product;
        });
    }

    @Override
    public String getManufacturer() {
        return vaccinationExtractor((CBORObject vacTag) -> {
            String manf = vacTag.get(CBORObject.FromObject(MANUFACTURER_TAG)).AsString();
            switch (manf)  {
            case "ORG-100030215":
                return "Biontech Manufacturing GmbH";
            }
            return manf;
        });
    }

    @Override
    public String getDose() {
        return vaccinationExtractor((CBORObject vacTag) -> {
            int dn = vacTag.get(CBORObject.FromObject(DOSENUM_TAG)).AsInt32();
            int sd = vacTag.get(CBORObject.FromObject(TOTALSERIALDOSE_TAG)).AsInt32();
            return String.format("%d/%d", dn, sd);
        });
    }

    @Override
    public String getDate() {
        return vaccinationExtractor((CBORObject vacTag) -> {
            return vacTag.get(CBORObject.FromObject(DATE_TAG)).AsString();
        });
    }

    @Override
    public String getCountry() {
        return vaccinationExtractor((CBORObject vacTag) -> {
            return vacTag.get(CBORObject.FromObject(VACCINATION_COUNTRY_TAG)).AsString();
        });
    }

    @Override
    public String getCertificatIssuer() {
        return vaccinationExtractor((CBORObject vacTag) -> {
            return vacTag.get(CBORObject.FromObject(ISSUER_TAG)).AsString();
        });

    }

    public CBORObject getNameTag() {
        return getGreenCert().get(CBORObject.FromObject(NAME_TAG));
    }

    @Override
    public String getSurname() {
        CBORObject nameTag = getNameTag();
        String fn = nameTag.get(CBORObject.FromObject(FAMILYNAME_TAG)).AsString();
        return fn;
    }

    @Override
    public String getForename() {
        CBORObject nameTag = getNameTag();
        String gn = nameTag.get(CBORObject.FromObject(GIVENNAME_TAG)).AsString();
        return gn;
    }

    @Override
    public String getName() {
        String fn = getSurname();
        String gn = getForename();
        return String.format("%s %s", gn, fn);
    }

    @Override
    public String getDateOfBirth() {
        return getGreenCert().get(CBORObject.FromObject(DATEOFBIRTH_TAG)).AsString();
    }

    @Override
    public String getVersion() {
        return getGreenCert().get(CBORObject.FromObject(VERSION_TAG)).AsString();
    }

    /**
     * @param bytecompressed compressed bytes
     * @return unziped bytes
     * @throws Exception if anything goes wrong
     */
    private byte[] decompress(byte[] bytecompressed) throws Exception {
        List<byte[]> buffers = new ArrayList<>();
        Inflater inflater = new Inflater();
        inflater.setInput(bytecompressed);
        int totalSize = 0;
        while (buffers.size() < 64) {   // limit this to a unlikely size
            byte[] buffer = new byte[BUFFER_SIZE];
            final int decompressedBytes = inflater.inflate(buffer);
            totalSize += decompressedBytes;
            buffers.add(buffer);
            if (inflater.finished()) {
                byte[] min = new byte[totalSize];
                int pos = 0;
                for (int i = 0; i < buffers.size(); ++i) {
                    byte[] buf = buffers.get(i);
                    int toCopy = (i < (buffers.size() - 1)) ? BUFFER_SIZE : decompressedBytes;
                    System.arraycopy(buf, 0, min, pos, toCopy);
                    pos += toCopy;
                }
                //System.out.format("Input %d out %d\n", bytecompressed.length, min.length);
                return min;
            }
        }
        throw new Exception(String.format("Too many buffers needed %d of %d each!!!\n", buffers.size(), BUFFER_SIZE));
    }

}
